---@class Ball
---@field x number
---@field y number
---@field xVelocity number
---@field yVelocity number
---@field z number
---@field flyBall boolean
---@field size number
---@field heldBy Fielder | nil
---@field xAnimator SimpleAnimator
---@field yAnimator SimpleAnimator
---@field sizeAnimator SimpleAnimator
---@field zAnimator SimpleAnimator
---@field private animatorLib pd_animator_lib
---@field private bounce thread Requires deltaSeconds on resume, returns ball height.
---@field private flyTimeMs number
Ball = {}

---@param animatorLib pd_animator_lib
---@return Ball
function Ball.new(animatorLib)
    return setmetatable({
        animatorLib = animatorLib,
        x = 400 --[[@as number]],
        y = 300 --[[@as number]],
        z = 0,
        flyBall = false,
        size = C.SmallestBallRadius,
        heldBy = nil --[[@type Runner | nil]],

        xAnimator = utils.staticAnimator(C.BallOffscreen),
        yAnimator = utils.staticAnimator(C.BallOffscreen),

        -- TODO? Replace these with a ballAnimatorZ?
        -- ...that might lose some of the magic of both. Compromise available? idk
        sizeAnimator = utils.staticAnimator(C.SmallestBallRadius),
        zAnimator = animatorLib.new(2000, -60, 0, utils.easingHill),
    }, { __index = Ball })
end

---@param fielder Fielder
function Ball:caughtBy(fielder)
    self.heldBy = fielder
    -- if self.flyBall then
    --     TODO: Signal an out if the ball hasn't touched the ground.
    -- end
end

function timeToFirstBounce(v)
    local h0 = 0.1 -- m/s
    -- local v = 10 -- m/s, current velocity
    local g = 10 -- m/s/s
    local t = 0 -- starting time
    local rho = 0.75 -- coefficient of restitution
    local tau = 0.10 -- contact time for bounce
    local hmax = h0 -- keep track of the maximum height
    local h = h0
    local hstop = 0.01 -- stop when bounce is less than 1 cm
    local freefall = true -- state: freefall or in contact
    local vmax = math.sqrt(2 * hmax * g)
    local dt = 0.1

    local ret = 0

    while hmax > hstop do
        if freefall then
            local hnew = h + v * dt - 0.5 * g * dt * dt
            if hnew < 0 then
                return ret
            else
                t = t + dt
                v = v - g * dt
                h = hnew
            end
        else
            t = t + tau
            vmax = vmax * rho
            v = vmax
            freefall = true
            h = 0
        end
        hmax = 0.5 * vmax * vmax / g
        ret = ret + dt
    end
    return 0
end

function bouncer(v, ball)
    printTable(ball)
    local startMs = playdate.getCurrentTimeMilliseconds()
    return function()
        local h0 = 0.1 -- m/s
        -- local v = 10 -- m/s, current velocity
        local g = 10 -- m/s/s
        local t = 0 -- starting time
        local rho = 0.60 -- coefficient of restitution
        local tau = 0.10 -- contact time for bounce
        local hmax = h0 -- keep track of the maximum height
        local h = h0
        local hstop = 0.01 -- stop when bounce is less than 1 cm
        local freefall = true -- state: freefall or in contact
        local t_last = -math.sqrt(2 * h0 / g) -- time we would have launched to get to h0 at t=0
        local vmax = math.sqrt(v * g)

        while hmax > hstop do
            local dt = coroutine.yield(h, not freefall)
            if freefall then
                local hnew = h + v * dt - 0.5 * g * dt * dt
                if hnew < 0 then
                    -- Bounced!
                    t = t_last + 2 * math.sqrt(2 * hmax / g)
                    freefall = false
                    t_last = t + tau
                    h = 0
                else
                    t = t + dt
                    v = v - g * dt
                    h = hnew
                end
            else
                t = t + tau
                vmax = vmax * rho
                v = vmax
                freefall = true
                h = 0
            end
            hmax = 0.5 * vmax * vmax / g
        end
        return 0
    end
end

---@param deltaSeconds number
function Ball:updatePosition(deltaSeconds)
    -- printTable({ x = self.x, y = self.y, z = self.z })
    if self.heldBy then
        utils.moveAtSpeedZ(self, deltaSeconds * 10, { x = self.heldBy.x, y = self.heldBy.y, z = C.GloveZ })
        -- self.x = self.heldBy.x
        -- self.y = self.heldBy.y
        -- self.z = C.GloveZ
        -- self.size = C.SmallestBallRadius
    else
        -- self.x = self.x + self.xVelocity
        -- self.x = self.xAnimator:currentValue()
        -- self.y = self.yAnimator:currentValue()
        -- self.z = self.zAnimator:currentValue()
        if self.bounce then
            local alive, z, justBounced = coroutine.resume(self.bounce, deltaSeconds)
            if alive then
                local lostVelMult = justBounced and 0.8 or 0.99

                self.xVelocity = self.xVelocity * (1 - (deltaSeconds * lostVelMult))
                self.yVelocity = self.yVelocity * (1 - (deltaSeconds * lostVelMult))

                self.x = self.x + (self.xVelocity * deltaSeconds)
                self.y = self.y + (self.yVelocity * deltaSeconds)

                self.z = z * 30
            else
                self.bounce = nil
            end
        end
        self.size = (math.max(0, 0.04 * (self.z - C.GloveZ))) + C.SmallestBallRadius
        -- print(self.size)
    end

    if self.z < 1 then
        self.flyBall = false
    end
end

---@alias DestinationAndFlightTime { destX: number, destY: number, flyTimeMs: number }

---@alias LaunchControls DestinationAndFlightTime

--- Launches the ball from its current position to the given destination.
---@param destX number
---@param destY number
---@param easingFunc EasingFunc
---@param freshHit boolean | nil
---@param flyTimeMs number | nil The angle away from parallel to the ground.
---  0 is straight forward, 90 is straight up, 180 is straight behind.
function Ball:launch(destX, destY, _, freshHit, _, power)
    if freshHit then
        self.flyBall = true
    end
    throwMeter:reset()
    self.heldBy = nil

    local flightDistance, x, y = utils.distanceBetween(self.x, self.y, destX, destY)
    local timeToBounce = timeToFirstBounce(10)

    -- if not flyTimeMs then
    --     flyTimeMs = flightDistance * C.DefaultLaunchPower
    -- end

    -- TODO? set a maxThrowDistance to limit throws by, instead
    power = power or 5
    self.xVelocity = -1.1 * x -- (x / -flightDistance) * power
    self.yVelocity = -1.1 * y -- (y / -flightDistance) * power
    printTable({ x = x, y = y })
    printTable({
        destX = destX,
        destY = destY,
        x = x,
        y = y,
    --     xVelocity = self.xVelocity,
    --     yVelocity = self.yVelocity,
    --     timeToBounce = timeToBounce,
    })
    -- -- TODO? Scale based on distance?
    -- self.yAnimator = self.animatorLib.new(flyTimeMs, self.y, destY, easingFunc)
    -- self.xAnimator = self.animatorLib.new(flyTimeMs, self.x, destX, easingFunc)
    -- self.zAnimator = self.animatorLib.new(flyTimeMs, flightDistance, C.GloveZ, utils.easingHill)

    self.bounce = coroutine.create(bouncer(10))
end
