--- @class Glove
--- @field z number

--- @class Fielder
--- @field glove Glove
--- @field catchEligible boolean
--- @field name string
--- @field x number
--- @field y number
--- @field target XyPair | nil
--- @field speed number
--- @field armStrength number

-- TODO: Run down baserunners in a pickle.

---@class Fielding
---@field fielders table<string, Fielder>
---@field fielderHoldingBall Fielder | nil
---@field private onFlyOut fun()
Fielding = {}

FielderDanceAnimator = playdate.graphics.animator.new(1, 10, 0, utils.easingHill)
FielderDanceAnimator.repeatCount = C.DanceBounceCount - 1

---@param name string
---@param speed number
---@return Fielder
local function newFielder(name, speed)
    return {
        name = name,
        speed = speed * C.FielderRunMult,
        catchEligible = true,
        armStrength = 10,
        glove = {
            z = C.GloveZ
        },
    }
end

function Fielding.new(onFlyOut)
    return setmetatable({
        fielders = {
            first = newFielder("First", 40),
            second = newFielder("Second", 40),
            shortstop = newFielder("Shortstop", 40),
            third = newFielder("Third", 40),
            pitcher = newFielder("Pitcher", 30),
            catcher = newFielder("Catcher", 35),
            left = newFielder("Left", 50),
            center = newFielder("Center", 50),
            right = newFielder("Right", 50),
        },
        onFlyOut = onFlyOut,
        ---@type Fielder | nil
        fielderHoldingBall = nil,
    }, { __index = Fielding })
end

--- Actually only benches the infield, because outfielders are far away!
---@param position XyPair
function Fielding:benchTo(position)
    self.fielders.first.target = position
    self.fielders.second.target = position
    self.fielders.shortstop.target = position
    self.fielders.third.target = position
    self.fielders.pitcher.target = position
    self.fielders.catcher.target = position
end

--- Resets the target positions of all fielders to their defaults (at their field positions).
---@param fromOffTheField XyPair | nil If provided, also sets all runners' current position to one centralized location.
function Fielding:resetFielderPositions(fromOffTheField)
    if fromOffTheField then
        for _, fielder in pairs(self.fielders) do
            fielder.x = fromOffTheField.x
            fielder.y = fromOffTheField.y
        end
    end

    self.fielders.first.target = utils.xy(C.Screen.W - 65, C.Screen.H * 0.48)
    self.fielders.second.target = utils.xy(C.Screen.W * 0.70, C.Screen.H * 0.30)
    self.fielders.shortstop.target = utils.xy(C.Screen.W * 0.30, C.Screen.H * 0.30)
    self.fielders.third.target = utils.xy(C.Screen.W * 0.1, C.Screen.H * 0.48)
    self.fielders.pitcher.target = utils.xy(C.PitcherStartPos.x, C.PitcherStartPos.y)
    self.fielders.catcher.target = utils.xy(C.Screen.W * 0.475, C.Screen.H * 0.92)
    self.fielders.left.target = utils.xy(C.Screen.W * -0.6, C.Screen.H * -0.1)
    self.fielders.center.target = utils.xy(C.Center.x, C.Screen.H * -0.4)
    self.fielders.right.target = utils.xy(C.Screen.W * 1.6, self.fielders.left.target.y)
end

---@param deltaSeconds number
---@param fielder Fielder
---@param ballPos Point3d
---@return boolean inCatchingRange
local function updateFielderPosition(deltaSeconds, fielder, ballPos)
    if fielder.target ~= nil then
        -- if fielder.name == "Left" then
        --     printTable({ target = fielder.target })
        -- end
        if fielder.target.z then
            if not utils.moveAtSpeedZ(fielder, fielder.speed * deltaSeconds, fielder.target) then
                if fielder.name == "Left" then
                    print("CLEAR LEFT'S 3D TARGET")
                end
                fielder.target = nil
            end
        else
            if not utils.moveAtSpeed(fielder, fielder.speed * deltaSeconds, fielder.target) then
                if fielder.name == "Left" then
                    print("CLEAR LEFT'S 2D TARGET")
                end
                fielder.target = nil
            end
        end
    end

    --local distance = utils.distanceBetweenZ(fielder.x, fielder.y, C.GloveZ, ballPos.x, ballPos.y, ballPos.z)
    if ballPos.z > C.GloveZ * 2 then
        return false
    end
    local distance = utils.distanceBetween(fielder.x, fielder.y, ballPos.x, ballPos.y)
    return distance < C.BallCatchHitbox
end

--- Selects the nearest fielder to move toward the given coordinates.
--- Other fielders should attempt to cover their bases
---@param ballDestX number
---@param ballDestY number
---@param ball Ball
function Fielding:haveSomeoneChase(ballDestX, ballDestY, ball)
    local chasingFielder = utils.getNearestOf(self.fielders, ballDestX, ballDestY)
    chasingFielder.target = ball
    -- local timer = playdate.timer.new(1000)
    -- timer.updateCallback = function()
    --     printTable(chasingFielder.target)
    -- end
    print("chasingFielder: " .. chasingFielder.name)
    printTable(ball)

    for _, base in ipairs(C.Bases) do
        local nearest = utils.getNearestOf(self.fielders, base.x, base.y, function(fielder)
            -- For now, skipping the pitcher because they're considered closer to 2B than second or shortstop
            return fielder ~= chasingFielder and fielder ~= self.fielders.pitcher
        end)
        nearest.target = base
    end
end

---@param ball Ball
---@param deltaSeconds number
---@return Fielder | nil fielderHoldingBall nil if no fielder is currently touching the ball
function Fielding:updateFielderPositions(ball, deltaSeconds)
    local fielderHoldingBall
    for _, fielder in pairs(self.fielders) do
        local inCatchingRange = updateFielderPosition(deltaSeconds, fielder, ball)
        -- if inCatchingRange then
        --     printTable({
        --         inCatchingRange = inCatchingRange,
        --         catchEligible = fielder.catchEligible,
        --         fielderName = fielder.name,
        --     })
        -- end
        if inCatchingRange and fielder.catchEligible then
            -- TODO: Base this catch on fielder skill?
            fielderHoldingBall = fielder
            if ball.flyBall then
                self.onFlyOut()
                ball.flyBall = false
            end
            ball:caughtBy(fielder)
        end
    end
    -- TODO: The need is growing for a distinction between touching the ball and holding the ball.
    -- Or, at least, fielders need to start *stopping* the ball when they make contact with it.
    -- Right now, a line-drive *through* first will be counted as an out.
    self.fielderHoldingBall = fielderHoldingBall
    return fielderHoldingBall
end

---@param fielder Fielder
function Fielding.markIneligible(fielder)
    fielder.catchEligible = false
    playdate.timer.new(500, function()
        fielder.catchEligible = true
    end)
end

function Fielding:markAllIneligible()
    for _, fielder in pairs(self.fielders) do
        fielder.catchEligible = false
    end
    playdate.timer.new(750, function()
        for _, fielder in pairs(self.fielders) do
            fielder.catchEligible = true
        end
    end)
end

-- TODO? Start moving target fielders close sooner?
---@param field Fielding
---@param targetBase Base
---@param ball { launch: LaunchBall }
---@param throwFlyMs number
local function userThrowToCoroutine(field, targetBase, ball, throwFlyMs)
    while field.fielderHoldingBall == nil do
        coroutine.yield()
    end

    local closestFielder = utils.getNearestOf(field.fielders, targetBase.x, targetBase.y, function(fielder)
        return fielder ~= field.fielderHoldingBall -- presumably, this is who will be doing the throwing
    end)

    closestFielder.target = targetBase
    ball:launch(targetBase.x, targetBase.y, playdate.easingFunctions.linear, false, throwFlyMs)
    Fielding.markIneligible(field.fielderHoldingBall)
end

--- Buffer in a fielder throw action.
---@param self table
---@param targetBase Base
---@param ball { launch: LaunchBall }
---@param throwFlyMs number
function Fielding:userThrowTo(targetBase, ball, throwFlyMs)
    local maxTryTimeMs = 5000
    actionQueue:upsert("userThrowTo", maxTryTimeMs, function()
        userThrowToCoroutine(self, targetBase, ball, throwFlyMs)
    end)
end

function Fielding.celebrate()
    FielderDanceAnimator:reset(C.DanceBounceMs)
end

-- luacheck: ignore
if not playdate or playdate.TEST_MODE then
    return { Fielding, newFielder }
end
