-- stylua: ignore start
import 'CoreLibs/animation.lua'
import 'CoreLibs/animator.lua'
import 'CoreLibs/easing.lua'
import 'CoreLibs/graphics.lua'
import 'CoreLibs/object.lua'
import 'CoreLibs/timer.lua'
import 'CoreLibs/ui.lua'
-- stylua: ignore end

--- @alias EasingFunc fun(number, number, number, number): number

--- @alias LaunchBall fun(
---   self: self,
---   destX: number,
---   destY: number,
---   easingFunc: EasingFunc,
---   freshHit: boolean | nil,
---   flyTimeMs: number | nil,
--- )

-- stylua: ignore start
import 'utils.lua'
import 'constants.lua'
import 'assets.lua'
import 'draw/player.lua'
import 'draw/overlay.lua'
import 'draw/fielder.lua'

import 'main-menu.lua'

import 'action-queue.lua'
import 'announcer.lua'
import 'ball.lua'
import 'baserunning.lua'
import 'dbg.lua'
import 'fielding.lua'
import 'graphics.lua'
import 'npc.lua'
-- stylua: ignore end

-- TODO: Customizable field structure. E.g. stands and ads etc.

local gfx <const>, C <const> = playdate.graphics, C

---@alias Team { score: number, benchPosition: XyPair }
---@type table<TeamId, Team>
local teams <const> = {
    home = {
        score = 0, -- TODO: Extract this last bit of global mutable state.
        benchPosition = utils.xy(C.Screen.W + 10, C.Center.y),
    },
    away = {
        score = 0,
        benchPosition = utils.xy(-10, C.Center.y),
    },
}

---@alias TeamId 'home' | 'away'

--- Well, maybe not "Settings", but passive state that probably won't change much, if at all, during a game.
---@class Settings
---@field finalInning number
---@field userTeam TeamId | nil
---@field awayTeamSprites SpriteCollection
---@field homeTeamSprites SpriteCollection

---@class MutableState
---@field deltaSeconds number
---@field ball Ball
---@field battingTeam TeamId
---@field catcherThrownBall boolean
---@field offenseState OffenseState
---@field inning number
---@field batBase XyPair
---@field batTip XyPair
---@field batAngleDeg number
-- TODO: Replace with timers, repeatedly reset, instead of constantly setting to 0
---@field secondsSinceLastRunnerMove number
---@field secondsSincePitchAllowed number
-- These are only sort-of global state. They are purely graphical,
-- but they need to be kept in sync with the rest of the globals.
---@field runnerBlipper Blipper
---@field battingTeamSprites SpriteCollection
---@field fieldingTeamSprites SpriteCollection

---@class Game
---@field private settings Settings
---@field private announcer Announcer
---@field private fielding Fielding
---@field private baserunning Baserunning
---@field private npc Npc
---@field private homeTeamBlipper Blipper
---@field private awayTeamBlipper Blipper
---@field private state MutableState
Game = {}

---@param settings Settings
---@param announcer Announcer | nil
---@param fielding Fielding | nil
---@param baserunning Baserunning | nil
---@param npc Npc | nil
---@param state MutableState | nil
---@return Game
function Game.new(settings, announcer, fielding, baserunning, npc, state)
    teams.away.score = 0
    teams.home.score = 0
    announcer = announcer or Announcer.new()
    settings.userTeam = "home" -- "away"

    local homeTeamBlipper = blipper.new(100, settings.homeTeamSprites.smiling, settings.homeTeamSprites.lowHat)
    local awayTeamBlipper = blipper.new(100, settings.awayTeamSprites.smiling, settings.awayTeamSprites.lowHat)
    local battingTeam = "away"
    local runnerBlipper = battingTeam == "away" and awayTeamBlipper or homeTeamBlipper
    local ball = Ball.new(gfx.animator)

    local o = setmetatable({
        settings = settings,
        announcer = announcer,
        homeTeamBlipper = homeTeamBlipper,
        awayTeamBlipper = awayTeamBlipper,
        state = state or {
            batBase = utils.xy(C.Center.x - 34, 215),
            batTip = utils.xy(0, 0),
            batAngleDeg = C.CrankOffsetDeg,
            deltaSeconds = 0,
            ball = ball,
            battingTeam = battingTeam,
            offenseState = C.Offense.batting,
            inning = 1,
            catcherThrownBall = true,
            secondsSinceLastRunnerMove = 0,
            secondsSincePitchAllowed = 0,
            battingTeamSprites = settings.awayTeamSprites,
            fieldingTeamSprites = settings.homeTeamSprites,
            runnerBlipper = runnerBlipper,
        },
    }, { __index = Game })

    o.fielding = fielding or Fielding.new(function()
        print("Fly out!")
    end)
    o.baserunning = baserunning or Baserunning.new(announcer, function()
        o:nextHalfInning()
    end)
    o.npc = npc or Npc.new(o.baserunning.runners, o.fielding.fielders)

    o.fielding:resetFielderPositions(teams.home.benchPosition)
    playdate.timer.new(3500, function()
        print("Start pitcher with ball")
        ball:launch(C.PitcherStartPos.x, C.PitcherStartPos.y, playdate.easingFunctions.linear, false, nil, 6)
    end)

    BootTune:play()
    BootTune:setFinishCallback(function()
        TinnyBackground:play()
    end)

    return o
end

---@alias SimpleAnimator { currentValue: fun(self): number; reset: fun(self, durationMs: number | nil) }
---@alias Pitch fun(ball: Ball): { x: SimpleAnimator, y: SimpleAnimator, z: SimpleAnimator | nil }

---@type Pitch[]
local Pitches <const> = {
    -- Fastball
    function()
        return {
            x = gfx.animator.new(0, C.PitchStartX, C.PitchStartX, playdate.easingFunctions.linear),
            y = gfx.animator.new(C.PitchFlyMs / 1.3, C.PitchStartY, C.PitchEndY, playdate.easingFunctions.linear),
        }
    end,
    -- Curve ball
    function()
        return {
            x = gfx.animator.new(C.PitchFlyMs, C.PitchStartX + 20, C.PitchStartX, utils.easingHill),
            y = gfx.animator.new(C.PitchFlyMs, C.PitchStartY, C.PitchEndY, playdate.easingFunctions.linear),
        }
    end,
    -- Slider
    function()
        return {
            x = gfx.animator.new(C.PitchFlyMs, C.PitchStartX - 20, C.PitchStartX, utils.easingHill),
            y = gfx.animator.new(C.PitchFlyMs, C.PitchStartY, C.PitchEndY, playdate.easingFunctions.linear),
        }
    end,
    -- Wobbbleball
    function(ball)
        return {
            x = {
                currentValue = function()
                    return C.PitchStartX + (10 * math.sin((ball.yAnimator:currentValue() - C.PitchStartY) / 10))
                end,
                reset = function() end,
            },
            y = gfx.animator.new(C.PitchFlyMs * 1.3, C.PitchStartY, C.PitchEndY, playdate.easingFunctions.linear),
        }
    end,
}

---@param teamId TeamId
---@return TeamId
local function getOppositeTeamId(teamId)
    if teamId == "home" then
        return "away"
    elseif teamId == "away" then
        return "home"
    else
        error("Unknown TeamId: " .. (teamId or "nil"))
    end
end

function Game:getBattingTeam()
    return teams[self.state.battingTeam] or error("Unknown battingTeam: " .. (self.state.battingTeam or "nil"))
end

function Game:getFieldingTeam()
    return teams[getOppositeTeamId(self.state.battingTeam)]
end

---@param side Side
---@return boolean userIsOnSide, boolean userIsOnOtherSide
function Game:userIsOn(side)
    if self.settings.userTeam == nil then
        -- Both teams are NPC-driven
        return false, false
    end
    local ret
    if self.settings.userTeam == self.state.battingTeam then
        ret = side == "offense"
    else
        ret = side == "defense"
    end
    return ret, not ret
end

---@param pitchFlyTimeMs number | nil
---@param pitchTypeIndex number | nil
function Game:pitch(pitchFlyTimeMs, pitchTypeIndex)
    Fielding.markIneligible(self.fielding.fielders.pitcher)
    self.state.ball:launch(C.PitchStartX, C.PitchEndY, nil, false, nil, 2000 / pitchFlyTimeMs)
    self.state.ball.heldBy = nil
    self.state.catcherThrownBall = false
    self.state.offenseState = C.Offense.batting

    -- local current = Pitches[pitchTypeIndex](self.state.ball)
    -- self.state.ball.xAnimator = current.x
    -- self.state.ball.yAnimator = current.y or Pitches[1](self.state.ball).y

    -- -- TODO: This would need to be sanely replaced in ball:launch() etc.
    -- -- if current.z then
    -- -- 	ball.zAnimator = current.z
    -- -- 	ball.zAnimator:reset()
    -- -- end

    -- if pitchFlyTimeMs then
    --     self.state.ball.xAnimator:reset(pitchFlyTimeMs)
    --     self.state.ball.yAnimator:reset(pitchFlyTimeMs)
    -- else
    --     self.state.ball.xAnimator:reset()
    --     self.state.ball.yAnimator:reset()
    -- end

    self.state.secondsSincePitchAllowed = 0
end

function Game:nextHalfInning()
    pitchTracker:reset()
    local gameOver = self.state.inning == self.settings.finalInning and teams.away.score ~= teams.home.score
    if not gameOver then
        Fielding.celebrate()
        self.state.secondsSinceLastRunnerMove = -7
        self.fielding:benchTo(self:getFieldingTeam().benchPosition)
        self.announcer:say("SWITCHING SIDES...")
    end

    if gameOver then
        self.announcer:say("AND THAT'S THE BALL GAME!")
    else
        self.fielding:resetFielderPositions()
        if self.state.battingTeam == teams.home then
            self.state.inning = self.state.inning + 1
        end
        self.state.battingTeam = getOppositeTeamId(self.state.battingTeam)
        playdate.timer.new(2000, function()
            if self.state.battingTeam == teams.home then
                self.state.battingTeamSprites = self.settings.homeTeamSprites
                self.state.runnerBlipper = self.homeTeamBlipper
                self.state.fieldingTeamSprites = self.settings.awayTeamSprites
            else
                self.state.battingTeamSprites = self.settings.awayTeamSprites
                self.state.fieldingTeamSprites = self.settings.homeTeamSprites
                self.state.runnerBlipper = self.awayTeamBlipper
            end
        end)
    end
end

---@param scoredRunCount number
function Game:score(scoredRunCount)
    local batting = self:getBattingTeam()
    batting.score = batting.score + scoredRunCount
    self.announcer:say("SCORE!")
end

---@param throwFlyMs number
---@return boolean didThrow
function Game:buttonControlledThrow(throwFlyMs, forbidThrowHome)
    local targetBase
    if playdate.buttonIsPressed(playdate.kButtonLeft) then
        targetBase = C.Bases[C.Third]
    elseif playdate.buttonIsPressed(playdate.kButtonUp) then
        targetBase = C.Bases[C.Second]
    elseif playdate.buttonIsPressed(playdate.kButtonRight) then
        targetBase = C.Bases[C.First]
    elseif not forbidThrowHome and playdate.buttonIsPressed(playdate.kButtonDown) then
        targetBase = C.Bases[C.Home]
    else
        return false
    end

    self.fielding:userThrowTo(targetBase, self.state.ball, throwFlyMs)
    self.state.secondsSinceLastRunnerMove = 0
    self.state.offenseState = C.Offense.running

    return true
end

function Game:nextBatter()
    self.state.secondsSincePitchAllowed = -3
    self.baserunning.batter = nil
    playdate.timer.new(2000, function()
        pitchTracker:reset()
        if not self.baserunning.batter then
            self.baserunning:pushNewBatter()
        end
    end)
end

function Game:walk()
    self.announcer:say("Walk!")
    -- TODO? Use self.baserunning:convertBatterToRunner()
    self.baserunning.batter.nextBase = C.Bases[C.First]
    self.baserunning.batter.prevBase = C.Bases[C.Home]
    self.state.offenseState = C.Offense.walking
    self.baserunning.batter = nil
    self.baserunning:updateForcedRunners()
    self:nextBatter()
end

function Game:strikeOut()
    local outBatter = self.baserunning.batter
    self.baserunning.batter = nil
    self.baserunning:outRunner(outBatter --[[@as Runner]], "Strike out!")
    self:nextBatter()
end

---@param batDeg number
function Game:updateBatting(batDeg, batSpeed)
    local batAngle = math.rad(batDeg)
    -- TODO: animate bat-flip or something
    self.state.batBase.x = self.baserunning.batter and (self.baserunning.batter.x + C.BatterHandPos.x) or 0
    self.state.batBase.y = self.baserunning.batter and (self.baserunning.batter.y + C.BatterHandPos.y) or 0
    self.state.batTip.x = self.state.batBase.x + (C.BatLength * math.sin(batAngle))
    self.state.batTip.y = self.state.batBase.y + (C.BatLength * math.cos(batAngle))

    if
        batSpeed > 0
        and utils.pointDirectlyUnderLine(
            self.state.ball.x,
            self.state.ball.y,
            self.state.batBase.x,
            self.state.batBase.y,
            self.state.batTip.x,
            self.state.batTip.y,
            C.Screen.H
        )
        and self.state.ball.y < 232
    then
        -- Hit!
        BatCrackReverb:play()
        self.state.offenseState = C.Offense.running
        -- local ballAngle = batAngle + math.rad(90)

        -- local mult = math.abs(batSpeed / 15)
        -- local ballVelX = mult * 10 * math.sin(ballAngle)
        -- local ballVelY = mult * 5 * math.cos(ballAngle)
        -- if ballVelY > 0 then
        --     ballVelX = ballVelX * -1
        --     ballVelY = ballVelY * -1
        -- end
        local ballDestX = self.fielding.fielders.left.x -- self.state.ball.x + (ballVelX * C.BattingPower)
        local ballDestY = self.fielding.fielders.left.y -- self.state.ball.y + (ballVelY * C.BattingPower)
        pitchTracker:reset()
        print("Hit ball!")
        self.state.ball:launch(ballDestX, ballDestY, playdate.easingFunctions.outQuint, true, 3000, batSpeed / 3)
        -- TODO? A dramatic eye-level view on a home-run could be sick.

        if utils.isFoulBall(ballDestX, ballDestY) then
            self.announcer:say("Foul ball!")
            pitchTracker.strikes = math.min(pitchTracker.strikes + 1, 2)
            -- TODO: Have a fielder chase for the fly-out
            return
        end

        self.baserunning:convertBatterToRunner()

        self.fielding:haveSomeoneChase(ballDestX, ballDestY, self.state.ball)
    end
end

---@param appliedSpeed number | fun(runner: Runner): number
---@return boolean someRunnerMoved
function Game:updateNonBatterRunners(appliedSpeed, forcedOnly)
    local runnerMoved, runnersScored =
        self.baserunning:updateNonBatterRunners(appliedSpeed, forcedOnly, self.state.deltaSeconds)
    if runnersScored ~= 0 then
        self:score(runnersScored)
    end
    return runnerMoved
end

---@param throwFly number
function Game:userPitch(throwFly)
    local aPressed = playdate.buttonIsPressed(playdate.kButtonA)
    local bPressed = playdate.buttonIsPressed(playdate.kButtonB)
    if not aPressed and not bPressed then
        self:pitch(throwFly, 1)
    elseif aPressed and not bPressed then
        self:pitch(throwFly, 2)
    elseif not aPressed and bPressed then
        self:pitch(throwFly, 3)
    elseif aPressed and bPressed then
        self:pitch(throwFly, 4)
    end
end

function Game:updateGameState()
    self.state.deltaSeconds = playdate.getElapsedTime() or 0
    playdate.resetElapsedTime()
    local crankChange = playdate.getCrankChange() --[[@as number]]
    local crankLimited = crankChange == 0 and 0 or (math.log(math.abs(crankChange)) * C.CrankPower)
    if crankChange < 0 then
        crankLimited = crankLimited * -1
    end

    self.state.ball:updatePosition(self.state.deltaSeconds)

    local userOnOffense, userOnDefense = self:userIsOn("offense")

    if userOnDefense then
        throwMeter:applyCharge(self.state.deltaSeconds, crankLimited)
    end

    if self.state.offenseState == C.Offense.batting then
        if self.state.ball.y < C.StrikeZoneStartY then
            pitchTracker.recordedPitchX = nil
        elseif not pitchTracker.recordedPitchX then
            pitchTracker.recordedPitchX = self.state.ball.x
        end

        local pitcher = self.fielding.fielders.pitcher
        if utils.distanceBetween(pitcher.x, pitcher.y, C.PitcherStartPos.x, C.PitcherStartPos.y) < C.BaseHitbox then
            self.state.secondsSincePitchAllowed = self.state.secondsSincePitchAllowed + self.state.deltaSeconds
        end

        if self.state.secondsSincePitchAllowed > C.ReturnToPitcherAfterSeconds and not self.state.catcherThrownBall then
            local outcome = pitchTracker:updatePitchCounts()
            if outcome == PitchOutcomes.StrikeOut then
                self:strikeOut()
            elseif outcome == PitchOutcomes.Walk then
                self:walk()
            end
            -- Catcher has the ball. Throw it back to the pitcher
            print("Catcher return ball to pitcher")
            self.state.ball:launch(C.PitcherStartPos.x, C.PitcherStartPos.y, playdate.easingFunctions.linear, false, 20)
            self.fielding:markAllIneligible()
            self.state.catcherThrownBall = true
        end

        local batSpeed
        if userOnOffense then
            self.state.batAngleDeg = (playdate.getCrankPosition() + C.CrankOffsetDeg) % 360
            batSpeed = crankLimited
        else
            self.state.batAngleDeg =
                self.npc:updateBatAngle(self.state.ball, self.state.catcherThrownBall, self.state.deltaSeconds)
            batSpeed = self.npc:batSpeed() * self.state.deltaSeconds
        end

        self:updateBatting(self.state.batAngleDeg, batSpeed)

        -- Walk batter to the plate
        -- TODO: Ensure batter can't be nil, here
        self.baserunning:updateRunner(self.baserunning.batter, nil, crankLimited, self.state.deltaSeconds)

        if self.state.secondsSincePitchAllowed > C.PitchAfterSeconds then
            if userOnDefense then
                local throwFly = throwMeter:readThrow()
                if throwFly and not self:buttonControlledThrow(throwFly, true) then
                    self:userPitch(throwFly)
                end
            else
                self:pitch(C.PitchFlyMs / self.npc:pitchSpeed(), math.random(#Pitches))
            end
        end
    elseif self.state.offenseState == C.Offense.running then
        local appliedSpeed = userOnOffense and crankLimited
            or function(runner)
                return self.npc:runningSpeed(runner, self.state.ball)
            end
        if self:updateNonBatterRunners(appliedSpeed) then
            self.state.secondsSinceLastRunnerMove = 0
        else
            self.state.secondsSinceLastRunnerMove = self.state.secondsSinceLastRunnerMove + self.state.deltaSeconds
            if self.state.secondsSinceLastRunnerMove > C.ResetFieldersAfterSeconds then
                -- End of play. Throw the ball back to the pitcher
                print("Return ball to pitcher")
                self.state.ball:launch(C.PitcherStartPos.x, C.PitcherStartPos.y, playdate.easingFunctions.linear, false, 20)
                -- This is ugly. Maybe Fielding should handle the return throw directly.
                self.fielding:markAllIneligible()
                self.fielding:resetFielderPositions()
                self.state.offenseState = C.Offense.batting
                -- TODO: Remove, or replace with nextBatter()
                if not self.baserunning.batter then
                    self.baserunning:pushNewBatter()
                end
            end
        end
    elseif self.state.offenseState == C.Offense.walking then
        if not self:updateNonBatterRunners(C.WalkedRunnerSpeed, true) then
            self.state.offenseState = C.Offense.batting
        end
    end

    local fielderHoldingBall = self.fielding:updateFielderPositions(self.state.ball, self.state.deltaSeconds)

    if userOnDefense then
        local throwFly = throwMeter:readThrow()
        if throwFly then
            self:buttonControlledThrow(throwFly)
        end
    end
    if fielderHoldingBall then
        local outedSomeRunner = self.baserunning:outEligibleRunners(fielderHoldingBall)
        if not userOnDefense and self.state.offenseState == C.Offense.running then
            self.npc:fielderAction(fielderHoldingBall, outedSomeRunner, self.state.ball)
        end
    end

    self.baserunning:walkAwayOutRunners(self.state.deltaSeconds)
    actionQueue:runWaiting(self.state.deltaSeconds)
end

-- TODO: Swappable update() for main menu, etc.

function Game:update()
    playdate.timer.updateTimers()
    gfx.animation.blinker.updateAll()
    self:updateGameState()
    local ball = self.state.ball

    gfx.clear()
    gfx.setColor(gfx.kColorBlack)

    local ballY = ball.y - (ball.z * 0.2)
    local offsetX, offsetY = getDrawOffset(ball.x, ballY)
    gfx.setDrawOffset(offsetX, offsetY)

    GrassBackground:draw(-400, -240)

    ---@type { y: number, drawAction: fun() }[]
    local characterDraws = {}
    function addDraw(y, drawAction)
        characterDraws[#characterDraws + 1] = { y = y, drawAction = drawAction }
    end

    local danceOffset = FielderDanceAnimator:currentValue()
    local ballIsHeld = false
    for _, fielder in pairs(self.fielding.fielders) do
        addDraw(fielder.y + danceOffset, function()
            ballIsHeld = drawFielder(
                self.state.fieldingTeamSprites,
                self.state.ball,
                fielder.x,
                fielder.y + danceOffset
            ) or ballIsHeld
        end)
    end

    local playerHeightOffset = 20
    -- TODO? Scale sprites down as y increases
    for _, runner in pairs(self.baserunning.runners) do
        addDraw(runner.y, function()
            if runner == self.baserunning.batter then
                if self.state.batAngleDeg > 50 and self.state.batAngleDeg < 200 then
                    self.state.battingTeamSprites.back:draw(runner.x, runner.y - playerHeightOffset)
                else
                    self.state.battingTeamSprites.smiling:draw(runner.x, runner.y - playerHeightOffset)
                end
            else
                -- TODO? Change blip speed depending on runner speed?
                self.state.runnerBlipper:draw(false, runner.x, runner.y - playerHeightOffset)
            end
        end)
    end

    table.sort(characterDraws, function(a, b)
        return a.y < b.y
    end)
    for _, character in pairs(characterDraws) do
        character.drawAction()
    end

    if self.state.offenseState == C.Offense.batting then
        gfx.setLineWidth(5)
        gfx.drawLine(self.state.batBase.x, self.state.batBase.y, self.state.batTip.x, self.state.batTip.y)
    end

    for _, runner in pairs(self.baserunning.outRunners) do
        self.state.battingTeamSprites.frowning:draw(runner.x, runner.y - playerHeightOffset)
    end

    if not ballIsHeld then
        gfx.setLineWidth(2)

        gfx.setColor(gfx.kColorWhite)
        gfx.fillCircleAtPoint(ball.x, ballY, ball.size)

        gfx.setColor(gfx.kColorBlack)
        gfx.drawCircleAtPoint(ball.x, ballY, ball.size)
    end

    gfx.setDrawOffset(0, 0)
    if math.abs(offsetX) > 10 or math.abs(offsetY) > 10 then
        drawMinimap(self.baserunning.runners, self.fielding.fielders)
    end
    drawScoreboard(0, C.Screen.H * 0.77, teams, self.baserunning.outs, self:getBattingTeam(), self.state.inning)
    drawBallsAndStrikes(290, C.Screen.H - 20, pitchTracker.balls, pitchTracker.strikes)
    self.announcer:draw(C.Center.x, 10)

    if playdate.isCrankDocked() then
        -- luacheck: ignore
        playdate.ui.crankIndicator:draw()
    end
end

playdate.display.setRefreshRate(50)
gfx.setBackgroundColor(gfx.kColorWhite)
playdate.setMenuImage(gfx.image.new("images/game/menu-image.png"))
playdate.getSystemMenu():addMenuItem("Restart game", function() end) -- TODO?

MainMenu.start(Game)

-- local b = coroutine.create(bouncer(10))
-- local x = 0
--
-- function playdate.update()
--     -- print(playdate.getFPS())
--     local deltaSeconds = playdate.getElapsedTime() or 0
--     playdate.resetElapsedTime()
--     local alive, z = coroutine.resume(b, deltaSeconds)
--     if alive then
--         z = z * 10
--         x = x + (deltaSeconds * 40)
--         gfx.setColor(gfx.kColorBlack)
--         gfx.drawCircleAtPoint(x, 240 - z, 5)
--     end
-- end
